#include <stdbool.h>
#include <stdint.h>

#define F_CPU 1000000UL

#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/power.h>
#include <avr/sleep.h>
#include <avr/wdt.h>

#define MOSI_PIN        0
#define MISO_PIN        1
#define SCK_PIN         2
#define POWER_PIN       3
#define IGNITION_PIN    4

#define SHUTDOWN_TIME_MS 1000UL * 60
#define GUARD_TIME_MS 1000UL * 5

#define ignition_on() (PINB & (1 << IGNITION_PIN))
#define power_on() (PINB & (1 << POWER_PIN))

enum power {
    off = 0,
    on
};

enum state {
    state_init = 0,
    state_on,
    state_shutdown,
    state_guard,
    state_off
};

volatile uint32_t tick;
uint32_t shutdown_expire, guard_expire;

enum state current_state;

void setup_gpio()
{
    DDRB = 0xff;
    PORTB = 0;
    DDRB &= ~(1 << IGNITION_PIN);
    PORTB |= (1 << IGNITION_PIN);   // enable pull-up
    PORTB &= ~(1 << MOSI_PIN);
    PORTB &= ~(1 << MISO_PIN);
    PORTB &= ~(1 << SCK_PIN);

    GIMSK |= (1 << PCIE);           // Enable Pin Change Interrupts
    PCMSK |= (1 << IGNITION_PIN);   // Use PB4 as interrupt pin
}

void setup_timer0()
{
    TCCR0A = (1 << WGM01);
    TCCR0B = (1 << CS01);   // prescaler = 8
    TIMSK |= (1 << OCIE0A); // enable compare match interrupt

    // ((1000000 / 8) / 1000) = 125
    OCR0A = 125 - 1;
}

void setup_watchdog()
{
    MCUSR &= ~(1 << WDRF);
    WDTCR = (1 << WDCE) | (1 << WDE);
    WDTCR = (1 << WDIE) | (1 << WDP2) | (1 << WDP1);
}

void switch_power(enum power power)
{
    if(power) {
        PORTB |= (1 << POWER_PIN);
    }
    else {
        PORTB &= ~(1 << POWER_PIN);
    }
}

void handle_state_init()
{
    if(ignition_on()) {
        switch_power(on);
        PORTB &= ~(1 << MOSI_PIN);
        PORTB &= ~(1 << MISO_PIN);
        PORTB &= ~(1 << SCK_PIN);
        current_state = state_on;
    } else {
        shutdown_expire = tick + SHUTDOWN_TIME_MS;
        PORTB |= (1 << MOSI_PIN);
        PORTB |= (1 << MISO_PIN);
        PORTB |= (1 << SCK_PIN);
        current_state = state_shutdown;
    }
}

void handle_state_on()
{
    if(!ignition_on()) {
        shutdown_expire = tick + SHUTDOWN_TIME_MS;
        PORTB |= (1 << MOSI_PIN);
        PORTB |= (1 << MISO_PIN);
        PORTB |= (1 << SCK_PIN);
        current_state = state_shutdown;
    }
}

void handle_state_shutdown()
{
    if(shutdown_expire < tick) {
        if(power_on()) {
            switch_power(off);
        }
        if(ignition_on()) {
            guard_expire = tick + GUARD_TIME_MS;
            current_state = state_guard;
        } else {
            current_state = state_off;
            sleep_enable();
            sleep_cpu();
            sleep_disable();
        }
    }
}

void handle_state_guard()
{
    if(guard_expire < tick) {
        if(ignition_on()) {
            switch_power(on);
            PORTB &= ~(1 << MOSI_PIN);
            PORTB &= ~(1 << MISO_PIN);
            PORTB &= ~(1 << SCK_PIN);
            current_state = state_on;
        } else {
            current_state = state_off;
            sleep_enable();
            sleep_cpu();
            sleep_disable();
        }
    }
}

void handle_state_off()
{
    if(ignition_on()) {
        switch_power(on);
        PORTB &= ~(1 << MOSI_PIN);
        PORTB &= ~(1 << MISO_PIN);
        PORTB &= ~(1 << SCK_PIN);
        current_state = state_on;
    }
}

int main(void)
{
    current_state = state_init;
    tick = 0;
    shutdown_expire = 0;
    guard_expire = 0;

    setup_gpio();
    setup_timer0();

    ADCSRA = 0;
    ADCSRB = 0;

    set_sleep_mode(SLEEP_MODE_PWR_DOWN);

    sei();

    while(1) {
        if(current_state == state_init) {
            handle_state_init();
        } else if(current_state == state_on) {
            handle_state_on();
        } else if(current_state == state_shutdown) {
            handle_state_shutdown();
        } else if(current_state == state_guard) {
            handle_state_guard();
        } else if(current_state == state_off) {
            handle_state_off();
        }
    }

    return 0;
}

ISR(TIM0_COMPA_vect)
{
    tick++;
}

// wake up by level interrupt on PB4
ISR(PCINT0_vect)
{
}

// wake up by watchdog interrupt
ISR(WDT_vect)
{
    WDTCR |= (1 << WDIE);
}

